// SingleInstance.cpp : implementation file
//
// Written By:
//
//	Erik Gawtry		August, 2005
//
// Usage:
//	On Startup (Constructor of CWinApp or base DLL's DLL_PROCESS_ATTACH):
//		OpenSiblingWindowHandles();
//
//	On Exit (Destructor of CWinApp or base DLL's DLL_PROCESS_DETACH):
//		CloseSiblingWindowHandles();
//
//
//	To Test in InitInstance():
//
//
//		CString sWindowID = _T("My Unique ID");		
//		HWND hWndOther = GetSiblingWindowHandle(sWindowID);
//		if( hWndOther != NULL )
//		{
//			TRACE(_T("Duplicate window found!\n"));
//			::SetWindowPos(hWndOther, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE);
//			SetFocus(hWndOther);
//			return FALSE;
//		}
//
//	When a new window is opened (like in MainFrame)
//
//	RegisterSiblingWindowHandle(_T("My Unique ID"), m_hWnd);
//
//
//  To unregister a window (in ExitInstance()):
//
//	RemoveSiblingWindowHandle(_T("My Unique ID"));
//
// Of course, replace all the _T("My Unique ID") tags with a unique ID for your app.
//
#include "stdafx.h"
#include "SingleInstance.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// Windows

 // Individual windows
typedef struct _tagSiblingWindowHandle
{
	WCHAR	swh_wszLabel[64];
	HWND	swh_hWnd;
} SIBLING_WINDOW_HANDLE;

 // Max number of windows stored
#define	SWH_MAX	256

 // Holder array for the window handles
typedef struct _tagSiblingWindowsHandles
{
	int nCount;
	SIBLING_WINDOW_HANDLE	SWH[SWH_MAX];
} SIBLING_WINDOW_HANDLES;

 // Get the list from shared memory
static bool GetSWHHandles(HANDLE &hSWH, SIBLING_WINDOW_HANDLES **pSWH)
{
	*pSWH = NULL;
	hSWH = ::CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, sizeof(SIBLING_WINDOW_HANDLES), _T("_SWH"));
	DWORD dwError = GetLastError();
	switch(dwError)
	{
	case ERROR_ALREADY_EXISTS:
		//TRACE(_T("CreateFileMapping: Found Handle\n"));
		*pSWH = (SIBLING_WINDOW_HANDLES *)MapViewOfFile(hSWH, FILE_MAP_ALL_ACCESS, 0, 0, 0);
		break;
	case ERROR_INVALID_HANDLE:
		//TRACE(_T("CreateFileMapping: InvalidHandle\n"));
		hSWH = NULL;
		break;
	default:
		//TRACE(_T("GetSWHHandles: Locking Header Memory\n"));
		*pSWH = (SIBLING_WINDOW_HANDLES *)MapViewOfFile(hSWH, FILE_MAP_ALL_ACCESS, 0, 0, 0);
		if( *pSWH != NULL )
		{
			//TRACE(_T("GetSWHHandles: Initializing Header Memory\n"));
			ZeroMemory(*pSWH, sizeof(SIBLING_WINDOW_HANDLES));
		}
		break;
	}
	return (hSWH!=NULL);
}

 // cleanup shared memory
static void CloseSWHHandles(SIBLING_WINDOW_HANDLES *pSWH)
{
	UnmapViewOfFile(pSWH);
}

 // Get the list and increment deletion counter
void OpenSiblingWindowHandles(void)
{
	HANDLE	hSWH;
	SIBLING_WINDOW_HANDLES	*pSWH;

	if( !GetSWHHandles(hSWH, &pSWH) ) return;

	++pSWH->nCount;
	CloseSWHHandles(pSWH);
}

 // Release the list and decrement deletion counter
void CloseSiblingWindowHandles(void)
{
	HANDLE	hSWH;
	SIBLING_WINDOW_HANDLES	*pSWH;

	if( !GetSWHHandles(hSWH, &pSWH) ) return;

	bool bCleanup = false;
	if( --pSWH->nCount == 0 )
		bCleanup = true;
	CloseSWHHandles(pSWH);

	 // Last one standing?
	if( bCleanup )
	{
		 // Close the shared memory
		::CloseHandle(hSWH);
	}
}

void RegisterSiblingWindowHandle(LPCTSTR lpLabel, HWND hWnd)
{
	HANDLE	hSWH;
	SIBLING_WINDOW_HANDLES	*pSWH;

	if( !GetSWHHandles(hSWH, &pSWH) ) return;

	WCHAR wszLabel[64];

	 // An update?
	for(int i = 0; i < SWH_MAX; ++i)
	{
#ifdef _UNICODE
		_tcsncpy(wszLabel, lpLabel, sizeof(pSWH->SWH[0].swh_wszLabel)/sizeof(WCHAR));
#else
		mbstowcs(wszLabel, lpLabel, sizeof(pSWH->SWH[0].swh_wszLabel)/sizeof(WCHAR));
#endif
		if( pSWH->SWH[i].swh_hWnd == hWnd )
		{
			wcscpy(pSWH->SWH[i].swh_wszLabel, wszLabel);
			CloseSWHHandles(pSWH);
			return;
		}
		 // Same as existing label?
		if( !wcscmp(pSWH->SWH[i].swh_wszLabel, wszLabel) )
		{
			 // Add a new one
			break;
		}
	}

	 // Or a new one?
	for(i = 0; i < SWH_MAX; ++i)
	{
		if( pSWH->SWH[i].swh_hWnd == NULL )
		{
			pSWH->SWH[i].swh_hWnd = hWnd;
#ifdef _UNICODE
			_tcsncpy(pSWH->SWH[i].swh_wszLabel, lpLabel, sizeof(pSWH->SWH[0].swh_wszLabel)/sizeof(WCHAR));
#else
			mbstowcs(pSWH->SWH[i].swh_wszLabel, lpLabel, sizeof(pSWH->SWH[0].swh_wszLabel)/sizeof(WCHAR));
#endif
			break;
		}
	}
	CloseSWHHandles(pSWH);
}

HWND GetSiblingWindowHandle(LPCTSTR lpLabel)
{
	HANDLE	hSWH;
	SIBLING_WINDOW_HANDLES	*pSWH;
	HWND hWndReturn = NULL;
	TCHAR	szLabel[64];

	if( !lpLabel || !GetSWHHandles(hSWH, &pSWH) ) return NULL;

	for(int i = 0; i < SWH_MAX; ++i)
	{
#ifdef _UNICODE
		_tcsncpy(szLabel, pSWH->SWH[i].swh_wszLabel, min(sizeof(pSWH->SWH[0].swh_wszLabel)/sizeof(WCHAR), sizeof(szLabel)/sizeof(TCHAR)));
#else
		wcstombs(szLabel, pSWH->SWH[i].swh_wszLabel, min(sizeof(pSWH->SWH[0].swh_wszLabel)/sizeof(WCHAR), sizeof(szLabel)/sizeof(TCHAR)));
#endif
		if( !_tcscmp(szLabel, lpLabel) )
		{
			if( ::IsWindow(pSWH->SWH[i].swh_hWnd) )
			{
				hWndReturn = pSWH->SWH[i].swh_hWnd;
				break;
			}
			else
			{
				 // This one is dead, try another
				pSWH->SWH[i].swh_hWnd = NULL;
				ZeroMemory(pSWH->SWH[i].swh_wszLabel, sizeof(pSWH->SWH[0].swh_wszLabel));
			}
		}
	}
	CloseSWHHandles(pSWH);
	return hWndReturn;
}

void RemoveSiblingWindowHandle(LPCTSTR lpLabel)
{
	HANDLE	hSWH;
	SIBLING_WINDOW_HANDLES	*pSWH;
	TCHAR	szLabel[64];

	if( !GetSWHHandles(hSWH, &pSWH) ) return;

	int nDel = -1;

	for(int i = 0; i < SWH_MAX; ++i)
	{
#ifdef _UNICODE
		_tcsncpy(szLabel, pSWH->SWH[i].swh_wszLabel, min(sizeof(pSWH->SWH[0].swh_wszLabel)/sizeof(WCHAR), sizeof(szLabel)/sizeof(TCHAR)));
#else
		wcstombs(szLabel, pSWH->SWH[i].swh_wszLabel, min(sizeof(pSWH->SWH[0].swh_wszLabel)/sizeof(WCHAR), sizeof(szLabel)/sizeof(TCHAR)));
#endif
		if( !_tcscmp(szLabel, lpLabel) )
		{
			if( ::IsWindow(pSWH->SWH[i].swh_hWnd) )
			{
				// Not this one, try the next
				nDel = i;
			}
			else
			{
				 // No window handle, so zap it
				pSWH->SWH[i].swh_hWnd = NULL;
				ZeroMemory(pSWH->SWH[i].swh_wszLabel, sizeof(pSWH->SWH[0].swh_wszLabel));
				goto Exit;
			}
		}
	}

	if( nDel >= 0 )
	{
		pSWH->SWH[nDel].swh_hWnd = NULL;
		ZeroMemory(pSWH->SWH[nDel].swh_wszLabel, sizeof(pSWH->SWH[0].swh_wszLabel));
	}
Exit:
	CloseSWHHandles(pSWH);
}

void RemoveSiblingWindowHandle(HWND hWnd)
{
	HANDLE	hSWH;
	SIBLING_WINDOW_HANDLES	*pSWH;

	if( !GetSWHHandles(hSWH, &pSWH) ) return;

	for(int i = 0; i < SWH_MAX; ++i)
	{
		if( pSWH->SWH[i].swh_hWnd == hWnd )
		{
			pSWH->SWH[i].swh_hWnd = NULL;
			ZeroMemory(pSWH->SWH[i].swh_wszLabel, sizeof(pSWH->SWH[0].swh_wszLabel));
			break;
		}
	}
	CloseSWHHandles(pSWH);
}

